import os
import numpy as np
import numba as nb
import casatools
import time

ms = casatools.ms()
tb = casatools.table()
me = casatools.measures()
qa = casatools.quanta()
ia = casatools.image()
im = casatools.imager()

@nb.njit(cache=True)
def populate_grid(uvw, stokesI, uvgrid, uvgrid_npt, deltau, deltav, npix):
    for ii in range(len(uvw[0])):
        uidx = int(np.round(uvw[0][ii]//deltau + npix//2))
        vidx = int(np.round(uvw[1][ii]//deltav + npix//2))

        uvgrid[uidx, vidx] += stokesI[ii]
        uvgrid_npt[uidx, vidx] += 1

    return uvgrid, uvgrid_npt


def mad(inpdat):
    """
    Calculate the STD via MAD for the input data

    Inputs:
    inpdat      Input numpy array

    Returns:
    std         Calculate the std via mad
    """

    med = np.median(inpdat)
    mad = np.median(np.abs(inpdat - med))

    return 1.4826 * mad


def msuvbin_to_uvgrid(ms, npix):
    tb.open(ms)
    uvw = tb.getcol('UVW')
    data = tb.getcol('DATA')
    tb.close()

    umin, umax = np.min(uvw[0]), np.max(uvw[0])
    vmin, vmax = np.min(uvw[1]), np.max(uvw[1])
    wmin, wmax = np.min(uvw[2]), np.max(uvw[2])

    uvals = np.linspace(umin, umax, npix)
    vvals = np.linspace(vmin, vmax, npix)

    deltau = uvals[1] - uvals[0]
    deltav = vvals[1] - vvals[0]

    uvgrid = np.zeros((npix, npix), dtype=np.complex128)
    uvgrid_npt = np.zeros((npix, npix), dtype=int)

    stokesI = 0.5*(data[0] + data[1])
    stokesI = np.squeeze(stokesI)

    uvgrid, uvgrid_npt = populate_grid(uvw, stokesI, uvgrid, uvgrid_npt, deltau, deltav, npix)
    # Average per uv cell
    idx = np.where(uvgrid_npt != 0)
    uvgrid[idx] = uvgrid[idx]/uvgrid_npt[idx]

    return uvgrid, uvgrid_npt



def calc_radial_profile(uvgrid, deltau, deltav, doplot=False):
    """
    Calculate the annular average of the uvgrid for every radius,
    and fit the 1D radial profile with a polynomial.
    """

    nbin = 10

    npixx, npixy = uvgrid.shape[0], uvgrid.shape[1]
    cx, cy = npixx//2, npixy//2

    print(f"npixx {npixx}, npixy {npixy}, centerx {cx} centery {cy}")

    # Generate grid of radii
    x = np.arange(npixx) - cx
    y = np.arange(npixy) - cy
    rad = np.sqrt(x**2 + y**2).astype(int)

    yy, xx = np.meshgrid(x, y)
    radgrid = np.sqrt(xx**2 + yy**2).astype(int)

    print(f"x shape {x.shape}, y shape {y.shape}, rad shape {rad.shape}")
    print(f"x min {x.min()}, x max {x.max()}")
    print(f"y min {y.min()}, y max {y.max()}")
    print(f"rad min {rad.min()}, rad max {rad.max()}")

    # Create log-spaced annuli to account for reducing UV coverage with radius
    # Minimum annulus is 5px
    annuli = np.logspace(np.log10(5), np.log10(rad.max()), nbin)
    annuli = np.round(annuli).astype(int)

    radial_mean = np.zeros(nbin)
    radial_mad = np.zeros(nbin)

    ann_min = 0
    ann_max = 0
    for idx, ann in enumerate(annuli):
        ann_max = ann_min + ann
        #print(f"ann_min {ann_min} ann_max {ann_max}")

        ridx = np.where((radgrid >= ann_min) & (radgrid < ann_max))
        #print(ridx)
        uvgrid_sel = uvgrid[ridx]
        uvgrid_sel = uvgrid_sel[np.abs(uvgrid_sel) != 0]

        if len(uvgrid_sel) == 0:
            radial_mean[idx] = 0
            radial_mad[idx] = 0.
        else:
            radial_mean[idx] = np.mean(uvgrid_sel)
            radial_mad[idx] = mad(uvgrid_sel)

        ann_min += ann

    if doplot:
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(figsize=(20,10))
        ax.plot(annuli * deltau, radial_mean, '-o')
        ax.fill_between(annuli*deltau, radial_mean-radial_mad, radial_mean+radial_mad, alpha=0.5)
        ax.set_xlabel('Radius (lambda)')
        ax.set_ylabel('Intensity (Jy)')
        ax.set_title('Radial Mean')
        plt.tight_layout()

        plt.savefig('radprof.jpg', bbox_inches='tight')

        fig, ax = plt.subplots(1, 1, figsize=(30,10))
        ax.imshow(np.abs(uvgrid), origin='lower', vmin=0, vmax=1e-3)
        ax.set_title('UV grid')
        plt.tight_layout()
        plt.savefig('uvgrid.jpg', bbox_inches='tight')


    print(f"pre flag non-zero points {np.count_nonzero(uvgrid)}")
    # Flag the data
    ann_min = 0
    ann_max = 0
    for idx, ann in enumerate(annuli):
        ann_max = ann_min + ann
        #print(f"ann_min {ann_min} ann_max {ann_max}")

        ridx = np.where((radgrid >= ann_min) & (radgrid < ann_max))
        uvgrid_sel = uvgrid[ridx]

        flagidx = np.where(np.abs(uvgrid_sel) > radial_mean[idx] + 5*radial_mad[idx])
        uvgrid_sel[flagidx] = 0

    print(f"post flag non-zero points {np.count_nonzero(uvgrid)}")
    return uvgrid, radial_mean, radial_mad


def flagViaBin_radial(binnedvis="", improveImp=True, sigma=5):
    tb.open(binnedvis, nomodify=False)
    msuvbinkey = tb.getkeyword('MSUVBIN')

    # msuvbinkey.keys()
    # Out[5]: dict_keys(['csys', 'nchan', 'npol', 'numvis', 'nx', 'ny', 'sumweight'])

    # in radian
    dra = msuvbinkey['csys']['direction0']['cdelt'][0]
    ddec = msuvbinkey['csys']['direction0']['cdelt'][1]

    nx = msuvbinkey['nx']
    ny = msuvbinkey['ny']

    # in radian
    ra_extent = dra*nx
    dec_extent = ddec*ny

    # in Lambda
    deltau = 1./ra_extent
    deltav = 1./dec_extent

    npol = msuvbinkey['npol']
    nchan = msuvbinkey['nchan']

    npoints = min(nx, ny)

    print(f"nx {nx}, ny {ny}, npol {npol}, nchan {nchan}, npoints {npoints}")
    print(f"dra {dra} ddec {ddec} ra_extent {ra_extent} dec_extent {dec_extent} deltau {deltau} deltav {deltav}")

    dat = tb.getcol('DATA')
    print(f"Data shape {dat.shape}")

    for pol in range(npol):
        for chan in range(nchan):

            dat = tb.getcolslice('DATA', [pol, chan], [pol, chan], [1, 1])
            flg = tb.getcolslice('FLAG', [pol, chan], [pol, chan], [1, 1])
            wgt = tb.getcolslice('WEIGHT_SPECTRUM', [pol, chan], [pol, chan], [1, 1])

            print(f"pol {pol}, chan {chan}")
            print(f"data shape {dat.shape}, flag shape {flg.shape}, weight shape {wgt.shape}")
            input()

            if dat.size == 0 or flg.size == 0 or wgt.size == 0:
                print(f"Zero size array read. Skipping.")
                continue

            dat_grid = dat[pol, chan, :].reshape([npoints, npoints])
            flg_grid = flg[pol, chan, :].reshape([npoints, npoints])
            wgt_grid = wgt[pol, chan, :].reshape([npoints, npoints])

            # Flag the data as necessary
            dat_grid = dat_grid * ~flg_grid

            uvgrid, radial_mean, radial_mad = calc_radial_profile(dat_grid, deltau, deltav, doplot=True)

            print(f"dat_grid shape {dat_grid.shape}, flg_grid shape {flg_grid.shape}, wgt_grid shape {wgt_grid.shape}")

            # Do the flagging and write back

            tb.putcolslice('DATA', dat, [pol, chan], [pol, chan])
            tb.putcolslice('FLAG', flg, [pol, chan], [pol, chan])
            tb.putcolslice('WEIGHT_SPECTRUM', wgt, [pol, chan], [pol, chan])

        tb.clearlocks()
        tb.done()


# @saveparams2last(multibackup=True)
def msuvbinflag(
        ######## binned ms
        binnedvis, #= 'thegridded_from_msuvbin.ms',
        ######## flagging algorithm option
        method, #= 'radial','RegionalMean','Gradient'
        sigma, #=0.5
        sizeRegion, #=20
        radius, #=1.0
        ignorPoint #= True
) -> None:

    if method == 'radial':
        tic = time.time()
        flagViaBin_radial(binnedvis=binnedvis, sigma=5)
        toc = time.time()
        print("MSuvbinflag radial method Running time :", toc - tic)
    elif method == 'regionalMean':
        print("No bueno")
        #tic = time.time()
        #flagViaBin_regionalMean(binnedvis=binnedvis, sizeRegion=20, sigma=5,ignorPoint=True)
        #toc = time.time()
        #print("MSuvbinflag other method Running time:", toc - tic)
    elif method == 'gradient':
        print("No bueno")
        #tic = time.time()
        #flagViaBin_gradient(binnedvis=binnedvis,  radius=1, sigma=0.3)
        #toc = time.time()
        #print("MSuvbinflag pead method Running time:", toc - tic)
    elif method == 'median':
        print("No bueno")
        #tic = time.time()
        #flagViaBin_median(binnedvis=binnedvis,  sigma=5)
        #toc = time.time()
        #print("MSuvbinflag median method Running time:", toc - tic)
    else:
        print("No bueno")
    return
